package g2g

import (
	"bytes"
	"encoding/json"
	"errors"
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"sync"

	"git.paulbsd.com/paulbsd/g2g/src/config"
)

// GetReposFromGitHub get starred repositories from github
func GetReposFromGitHub(config *config.Config) ([]GitHubRepo, error) {
	var repopartiallist []GitHubRepo
	var repofulllist []GitHubRepo
	var pagenum = 1

	fmt.Println("Getting GitHub starred repos")
	for {
		url := fmt.Sprintf("https://api.github.com/users/%s/starred?page=%d", config.GitHubAuthUsername, pagenum)
		fmt.Println(url)

		resp, err := GetGitHubResponse(config, url)
		if err != nil {
			return nil, err
		}

		respbody, err := ioutil.ReadAll(resp.Body)
		if err != nil {
			return nil, err
		}

		err = json.Unmarshal(respbody, &repopartiallist)
		if err != nil {
			return nil, err
		}

		if len(repopartiallist) == 0 {
			break
		}

		for _, elem := range repopartiallist {
			repofulllist = append(repofulllist, elem)
		}
		pagenum++
	}

	fmt.Printf("%d repositories fetched from Github\n", len(repofulllist))
	return repofulllist, nil
}

// GetGitHubResponse fetchs starred repositories on GitHub
func GetGitHubResponse(config *config.Config, url string) (*http.Response, error) {
	req, err := http.NewRequest("GET", url, nil)
	if err != nil {
		return nil, err
	}

	req.Header.Add("Content-Type", "application/x-www-form-urlencoded")
	req.SetBasicAuth(config.GitHubAuthUsername, config.GitHubAuthPassword)

	client := &http.Client{}
	resp, err := client.Do(req)
	if err != nil {
		return nil, err
	}

	if resp.StatusCode != http.StatusOK {
		return nil, fmt.Errorf("error on status code %s", resp.Status)
	}

	return resp, nil
}

// CheckGiteaExistingRepo checks if there's an existing repository in gitea
func CheckGiteaExistingRepo(config *config.Config, repo GitHubRepo) (bool, error) {
	var isExists bool

	gitearepourl := fmt.Sprintf(config.GiteaRepoURLTmpl, config.GiteaDestUsername, repo.Name)

	req, err := http.NewRequest("GET", gitearepourl, nil)
	if err != nil {
		return false, err
	}
	req.Header.Set("Authorization", config.GiteaAuthToken)

	client := &http.Client{}
	resp, err := client.Do(req)
	if err != nil {
		return false, err
	}

	if resp.StatusCode == http.StatusOK {
		isExists = true
	} else if resp.StatusCode == http.StatusNotFound {
		isExists = false
	} else {
		return false, fmt.Errorf("can't determine error, cancelling, error %s in gitea webservice", resp.Status)
	}

	return isExists, nil
}

// GetGiteaUserUID get the logued user identifier
func GetGiteaUserUID(config *config.Config) error {
	var giteaorg GiteaOrg

	gitearepourl := fmt.Sprintf(config.GiteaOrgsURLTmpl, config.GiteaDestUsername)

	req, err := http.NewRequest("GET", gitearepourl, nil)
	if err != nil {
		return err
	}

	req.Header.Set("Authorization", config.GiteaAuthToken)

	client := &http.Client{}
	resp, err := client.Do(req)
	if err != nil {
		return err
	}

	if resp.StatusCode != 200 {
		err = errors.New("error invoking gitea webservice")
		return err
	}

	respbody, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		return err
	}

	err = json.Unmarshal(respbody, &giteaorg)
	if err != nil {
		err = errors.New("failed to parse user ID from gitea webservice, check auth")
		return err
	}

	config.GiteaUID = giteaorg.ID

	return nil
}

// RunMigration run threads for migration
func RunMigration(config *config.Config, repolist []GitHubRepo) {
	repochan := make(chan GitHubRepo)
	done := make(chan bool)

	var wg sync.WaitGroup
	for thr := range make([]int, config.G2gThreads) {
		go MigrateReposToGitea(config, &wg, repochan, done, thr)
	}

	for _, repo := range repolist {
		repochan <- repo
	}
	close(repochan)
	<-done
	wg.Wait()
}

// MigrateReposToGitea migrates input repositories to gitea
func MigrateReposToGitea(config *config.Config, wg *sync.WaitGroup, jobs chan GitHubRepo, done chan bool, thr int) error {
	wg.Add(1)

	for {
		elem, more := <-jobs
		if more {
			existingrepo, err := CheckGiteaExistingRepo(config, elem)
			if err != nil {
				return err
			}
			if !existingrepo {
				err = MigrateRepo(elem, config)
				if err != nil {
					log.Println(err)
					continue
				}
			} else {
				fmt.Printf("Not migrating, %s exists in gitea\n", elem.Name)
			}
		} else {
			fmt.Printf("All repo migrated on thread num %d\n", thr)
			wg.Done()
			done <- true
			return nil
		}
	}
}

func MigrateRepo(elem GitHubRepo, config *config.Config) (err error) {
	client := &http.Client{}
	repo := GiteaRepo{
		UID:       config.GiteaUID,
		RepoName:  elem.Name,
		Mirror:    true,
		CloneAddr: elem.CloneURL}
	jsondata, err := json.Marshal(repo)
	if err != nil {
		return err
	}
	req, err := http.NewRequest("POST", config.GiteaMigrateURL, bytes.NewBufferString(string(jsondata)))
	if err != nil {
		return err
	}

	req.Header.Add("Content-Type", "application/json")
	req.Header.Set("Authorization", config.GiteaAuthToken)

	client.Timeout = config.G2gRequestTimeout

	fmt.Printf("Migrating repo %s to gitea rest api\n", elem.Name)
	resp, err := client.Do(req)
	if err != nil {
		return err
	}

	if resp.StatusCode != 201 {
		err = fmt.Errorf("error when migrating repo %s to gitea with status code %d on gitea webservice", elem.Name, resp.StatusCode)
		log.Println(err)
	}
	return err
}

// GitHubRepo githubrepo struct
type GitHubRepo struct {
	Name     string `json:"name"`
	CloneURL string `json:"clone_url"`
}

// GiteaOrg gitearepo struct
type GiteaOrg struct {
	ID       int    `json:"id"`
	Username string `json:"username"`
}

// GiteaRepo defines the repo to migrate
type GiteaRepo struct {
	UID       int    `json:"uid"`
	RepoName  string `json:"repo_name"`
	Mirror    bool   `json:"mirror"`
	CloneAddr string `json:"clone_addr"`
}
